# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from neuroshard.protos import neuroshard_pb2 as neuroshard__pb2

GRPC_GENERATED_VERSION = '1.76.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in neuroshard_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class NeuroShardServiceStub(object):
    """Service definition for Node-to-Node and Client-to-Node communication
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StreamInference = channel.stream_stream(
                '/neuroshard.NeuroShardService/StreamInference',
                request_serializer=neuroshard__pb2.InferenceRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.InferenceResponse.FromString,
                _registered_method=True)
        self.UnaryInference = channel.unary_unary(
                '/neuroshard.NeuroShardService/UnaryInference',
                request_serializer=neuroshard__pb2.InferenceRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.InferenceResponse.FromString,
                _registered_method=True)
        self.GetWeights = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetWeights',
                request_serializer=neuroshard__pb2.WeightRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.WeightResponse.FromString,
                _registered_method=True)
        self.GossipProof = channel.unary_unary(
                '/neuroshard.NeuroShardService/GossipProof',
                request_serializer=neuroshard__pb2.GossipProofRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.GossipProofResponse.FromString,
                _registered_method=True)
        self.GossipTransaction = channel.unary_unary(
                '/neuroshard.NeuroShardService/GossipTransaction',
                request_serializer=neuroshard__pb2.GossipTransactionRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.GossipTransactionResponse.FromString,
                _registered_method=True)
        self.GossipStake = channel.unary_unary(
                '/neuroshard.NeuroShardService/GossipStake',
                request_serializer=neuroshard__pb2.GossipStakeRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.GossipStakeResponse.FromString,
                _registered_method=True)
        self.RequestProofValidation = channel.unary_unary(
                '/neuroshard.NeuroShardService/RequestProofValidation',
                request_serializer=neuroshard__pb2.ProofValidationRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.ProofValidationResponse.FromString,
                _registered_method=True)
        self.GossipValidationVote = channel.unary_unary(
                '/neuroshard.NeuroShardService/GossipValidationVote',
                request_serializer=neuroshard__pb2.ValidationVoteRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.ValidationVoteResponse.FromString,
                _registered_method=True)
        self.GossipGradient = channel.unary_unary(
                '/neuroshard.NeuroShardService/GossipGradient',
                request_serializer=neuroshard__pb2.GossipGradientRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.GossipGradientResponse.FromString,
                _registered_method=True)
        self.GetCheckpoint = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetCheckpoint',
                request_serializer=neuroshard__pb2.GetCheckpointRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.GetCheckpointResponse.FromString,
                _registered_method=True)
        self.GetCheckpointInfo = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetCheckpointInfo',
                request_serializer=neuroshard__pb2.GetCheckpointInfoRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.GetCheckpointInfoResponse.FromString,
                _registered_method=True)
        self.PipelineForward = channel.unary_unary(
                '/neuroshard.NeuroShardService/PipelineForward',
                request_serializer=neuroshard__pb2.PipelineForwardRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.PipelineForwardResponse.FromString,
                _registered_method=True)
        self.PipelineBackward = channel.unary_unary(
                '/neuroshard.NeuroShardService/PipelineBackward',
                request_serializer=neuroshard__pb2.PipelineBackwardRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.PipelineBackwardResponse.FromString,
                _registered_method=True)
        self.GetShardInfo = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetShardInfo',
                request_serializer=neuroshard__pb2.GetShardInfoRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.GetShardInfoResponse.FromString,
                _registered_method=True)
        self.GetShardChunk = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetShardChunk',
                request_serializer=neuroshard__pb2.GetShardChunkRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.GetShardChunkResponse.FromString,
                _registered_method=True)
        self.DHTPing = channel.unary_unary(
                '/neuroshard.NeuroShardService/DHTPing',
                request_serializer=neuroshard__pb2.DHTPingRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.DHTPingResponse.FromString,
                _registered_method=True)
        self.DHTStore = channel.unary_unary(
                '/neuroshard.NeuroShardService/DHTStore',
                request_serializer=neuroshard__pb2.DHTStoreRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.DHTStoreResponse.FromString,
                _registered_method=True)
        self.DHTFindNode = channel.unary_unary(
                '/neuroshard.NeuroShardService/DHTFindNode',
                request_serializer=neuroshard__pb2.DHTFindNodeRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.DHTFindNodeResponse.FromString,
                _registered_method=True)
        self.DHTFindValue = channel.unary_unary(
                '/neuroshard.NeuroShardService/DHTFindValue',
                request_serializer=neuroshard__pb2.DHTFindValueRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.DHTFindValueResponse.FromString,
                _registered_method=True)
        self.TensorExchange = channel.unary_unary(
                '/neuroshard.NeuroShardService/TensorExchange',
                request_serializer=neuroshard__pb2.TensorExchangeRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.TensorExchangeResponse.FromString,
                _registered_method=True)
        self.SendPartialResult = channel.unary_unary(
                '/neuroshard.NeuroShardService/SendPartialResult',
                request_serializer=neuroshard__pb2.PartialResultRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.PartialResultResponse.FromString,
                _registered_method=True)
        self.AnnounceTensorShard = channel.unary_unary(
                '/neuroshard.NeuroShardService/AnnounceTensorShard',
                request_serializer=neuroshard__pb2.AnnounceShardRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.AnnounceShardResponse.FromString,
                _registered_method=True)
        self.FindTensorShardPeers = channel.unary_unary(
                '/neuroshard.NeuroShardService/FindTensorShardPeers',
                request_serializer=neuroshard__pb2.FindShardPeersRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.FindShardPeersResponse.FromString,
                _registered_method=True)
        self.ListModels = channel.unary_unary(
                '/neuroshard.NeuroShardService/ListModels',
                request_serializer=neuroshard__pb2.ListModelsRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.ListModelsResponse.FromString,
                _registered_method=True)
        self.GetModelStatus = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetModelStatus',
                request_serializer=neuroshard__pb2.GetModelStatusRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.GetModelStatusResponse.FromString,
                _registered_method=True)
        self.SwarmForward = channel.unary_unary(
                '/neuroshard.NeuroShardService/SwarmForward',
                request_serializer=neuroshard__pb2.SwarmForwardRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.SwarmForwardResponse.FromString,
                _registered_method=True)
        self.GetSwarmStatus = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetSwarmStatus',
                request_serializer=neuroshard__pb2.SwarmStatusRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.SwarmStatusResponse.FromString,
                _registered_method=True)
        self.UpdatePeerCapacity = channel.unary_unary(
                '/neuroshard.NeuroShardService/UpdatePeerCapacity',
                request_serializer=neuroshard__pb2.UpdatePeerCapacityRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.UpdatePeerCapacityResponse.FromString,
                _registered_method=True)
        self.AnnounceEpoch = channel.unary_unary(
                '/neuroshard.NeuroShardService/AnnounceEpoch',
                request_serializer=neuroshard__pb2.EpochAnnouncementRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.EpochAnnouncementResponse.FromString,
                _registered_method=True)
        self.GetEpoch = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetEpoch',
                request_serializer=neuroshard__pb2.EpochRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.EpochResponse.FromString,
                _registered_method=True)
        self.GetLatestEpoch = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetLatestEpoch',
                request_serializer=neuroshard__pb2.Empty.SerializeToString,
                response_deserializer=neuroshard__pb2.LatestEpochResponse.FromString,
                _registered_method=True)
        self.SpotCheckChallenge = channel.unary_unary(
                '/neuroshard.NeuroShardService/SpotCheckChallenge',
                request_serializer=neuroshard__pb2.SpotCheckChallengeRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.SpotCheckChallengeResponse.FromString,
                _registered_method=True)
        self.ExpertForward = channel.unary_unary(
                '/neuroshard.NeuroShardService/ExpertForward',
                request_serializer=neuroshard__pb2.ExpertForwardRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.ExpertForwardResponse.FromString,
                _registered_method=True)
        self.GetExpertStatus = channel.unary_unary(
                '/neuroshard.NeuroShardService/GetExpertStatus',
                request_serializer=neuroshard__pb2.ExpertStatusRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.ExpertStatusResponse.FromString,
                _registered_method=True)
        self.AnnounceExpert = channel.unary_unary(
                '/neuroshard.NeuroShardService/AnnounceExpert',
                request_serializer=neuroshard__pb2.AnnounceExpertRequest.SerializeToString,
                response_deserializer=neuroshard__pb2.AnnounceExpertResponse.FromString,
                _registered_method=True)


class NeuroShardServiceServicer(object):
    """Service definition for Node-to-Node and Client-to-Node communication
    """

    def StreamInference(self, request_iterator, context):
        """Streaming inference: Client/Node sends a stream of requests, receives a stream of tokens/updates
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UnaryInference(self, request, context):
        """Simple unary call for single-step (legacy/fallback)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetWeights(self, request, context):
        """Gossip weights for training
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GossipProof(self, request, context):
        """Gossip Proof of Uptime
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GossipTransaction(self, request, context):
        """Gossip Transaction
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GossipStake(self, request, context):
        """Gossip Stake Update
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RequestProofValidation(self, request, context):
        """Request Proof Validation from Validators
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GossipValidationVote(self, request, context):
        """Gossip Validation Vote
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GossipGradient(self, request, context):
        """--- Distributed Training RPCs ---

        Gossip gradients for distributed training
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCheckpoint(self, request, context):
        """Request checkpoint from peer
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCheckpointInfo(self, request, context):
        """Get checkpoint info (version, hash) without downloading
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PipelineForward(self, request, context):
        """--- Pipeline Parallelism RPCs ---

        Forward hidden states through this node's layers
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PipelineBackward(self, request, context):
        """Backward pass: propagate gradients back to previous node
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetShardInfo(self, request, context):
        """Get shard info from this node
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetShardChunk(self, request, context):
        """--- Data Swarm RPCs (P2P Dataset) ---

        Request a chunk of a data shard
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DHTPing(self, request, context):
        """--- DHT RPCs ---
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DHTStore(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DHTFindNode(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DHTFindValue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TensorExchange(self, request, context):
        """--- Phase 4: Tensor Parallelism RPCs ---

        Exchange tensor chunks during ring all-reduce
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendPartialResult(self, request, context):
        """Send partial results for async aggregation
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AnnounceTensorShard(self, request, context):
        """Announce tensor shard availability
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindTensorShardPeers(self, request, context):
        """Find peer shards for all-reduce coordination
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListModels(self, request, context):
        """--- Phase 4: Model Registry RPCs ---

        List available models in network
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetModelStatus(self, request, context):
        """Get status of a specific model
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SwarmForward(self, request, context):
        """--- Swarm Routing RPCs (Phase 2) ---

        Async activation forward (non-blocking, buffers locally)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSwarmStatus(self, request, context):
        """Get swarm node status (buffer fill rates, capacity)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdatePeerCapacity(self, request, context):
        """Update peer capacity (TCP fallback for UDP heartbeat)
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AnnounceEpoch(self, request, context):
        """--- Chained Epoch System RPCs ---

        Announce a finalized epoch to peers
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEpoch(self, request, context):
        """Request a specific epoch from peer
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLatestEpoch(self, request, context):
        """Get latest finalized epoch ID
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SpotCheckChallenge(self, request, context):
        """Submit spot-check challenge for gradient commitment
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExpertForward(self, request, context):
        """--- Mixture of Experts (MoE) RPCs ---

        Forward activations to a remote expert
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetExpertStatus(self, request, context):
        """Get expert status for routing decisions
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AnnounceExpert(self, request, context):
        """Announce expert availability to network
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_NeuroShardServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'StreamInference': grpc.stream_stream_rpc_method_handler(
                    servicer.StreamInference,
                    request_deserializer=neuroshard__pb2.InferenceRequest.FromString,
                    response_serializer=neuroshard__pb2.InferenceResponse.SerializeToString,
            ),
            'UnaryInference': grpc.unary_unary_rpc_method_handler(
                    servicer.UnaryInference,
                    request_deserializer=neuroshard__pb2.InferenceRequest.FromString,
                    response_serializer=neuroshard__pb2.InferenceResponse.SerializeToString,
            ),
            'GetWeights': grpc.unary_unary_rpc_method_handler(
                    servicer.GetWeights,
                    request_deserializer=neuroshard__pb2.WeightRequest.FromString,
                    response_serializer=neuroshard__pb2.WeightResponse.SerializeToString,
            ),
            'GossipProof': grpc.unary_unary_rpc_method_handler(
                    servicer.GossipProof,
                    request_deserializer=neuroshard__pb2.GossipProofRequest.FromString,
                    response_serializer=neuroshard__pb2.GossipProofResponse.SerializeToString,
            ),
            'GossipTransaction': grpc.unary_unary_rpc_method_handler(
                    servicer.GossipTransaction,
                    request_deserializer=neuroshard__pb2.GossipTransactionRequest.FromString,
                    response_serializer=neuroshard__pb2.GossipTransactionResponse.SerializeToString,
            ),
            'GossipStake': grpc.unary_unary_rpc_method_handler(
                    servicer.GossipStake,
                    request_deserializer=neuroshard__pb2.GossipStakeRequest.FromString,
                    response_serializer=neuroshard__pb2.GossipStakeResponse.SerializeToString,
            ),
            'RequestProofValidation': grpc.unary_unary_rpc_method_handler(
                    servicer.RequestProofValidation,
                    request_deserializer=neuroshard__pb2.ProofValidationRequest.FromString,
                    response_serializer=neuroshard__pb2.ProofValidationResponse.SerializeToString,
            ),
            'GossipValidationVote': grpc.unary_unary_rpc_method_handler(
                    servicer.GossipValidationVote,
                    request_deserializer=neuroshard__pb2.ValidationVoteRequest.FromString,
                    response_serializer=neuroshard__pb2.ValidationVoteResponse.SerializeToString,
            ),
            'GossipGradient': grpc.unary_unary_rpc_method_handler(
                    servicer.GossipGradient,
                    request_deserializer=neuroshard__pb2.GossipGradientRequest.FromString,
                    response_serializer=neuroshard__pb2.GossipGradientResponse.SerializeToString,
            ),
            'GetCheckpoint': grpc.unary_unary_rpc_method_handler(
                    servicer.GetCheckpoint,
                    request_deserializer=neuroshard__pb2.GetCheckpointRequest.FromString,
                    response_serializer=neuroshard__pb2.GetCheckpointResponse.SerializeToString,
            ),
            'GetCheckpointInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetCheckpointInfo,
                    request_deserializer=neuroshard__pb2.GetCheckpointInfoRequest.FromString,
                    response_serializer=neuroshard__pb2.GetCheckpointInfoResponse.SerializeToString,
            ),
            'PipelineForward': grpc.unary_unary_rpc_method_handler(
                    servicer.PipelineForward,
                    request_deserializer=neuroshard__pb2.PipelineForwardRequest.FromString,
                    response_serializer=neuroshard__pb2.PipelineForwardResponse.SerializeToString,
            ),
            'PipelineBackward': grpc.unary_unary_rpc_method_handler(
                    servicer.PipelineBackward,
                    request_deserializer=neuroshard__pb2.PipelineBackwardRequest.FromString,
                    response_serializer=neuroshard__pb2.PipelineBackwardResponse.SerializeToString,
            ),
            'GetShardInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetShardInfo,
                    request_deserializer=neuroshard__pb2.GetShardInfoRequest.FromString,
                    response_serializer=neuroshard__pb2.GetShardInfoResponse.SerializeToString,
            ),
            'GetShardChunk': grpc.unary_unary_rpc_method_handler(
                    servicer.GetShardChunk,
                    request_deserializer=neuroshard__pb2.GetShardChunkRequest.FromString,
                    response_serializer=neuroshard__pb2.GetShardChunkResponse.SerializeToString,
            ),
            'DHTPing': grpc.unary_unary_rpc_method_handler(
                    servicer.DHTPing,
                    request_deserializer=neuroshard__pb2.DHTPingRequest.FromString,
                    response_serializer=neuroshard__pb2.DHTPingResponse.SerializeToString,
            ),
            'DHTStore': grpc.unary_unary_rpc_method_handler(
                    servicer.DHTStore,
                    request_deserializer=neuroshard__pb2.DHTStoreRequest.FromString,
                    response_serializer=neuroshard__pb2.DHTStoreResponse.SerializeToString,
            ),
            'DHTFindNode': grpc.unary_unary_rpc_method_handler(
                    servicer.DHTFindNode,
                    request_deserializer=neuroshard__pb2.DHTFindNodeRequest.FromString,
                    response_serializer=neuroshard__pb2.DHTFindNodeResponse.SerializeToString,
            ),
            'DHTFindValue': grpc.unary_unary_rpc_method_handler(
                    servicer.DHTFindValue,
                    request_deserializer=neuroshard__pb2.DHTFindValueRequest.FromString,
                    response_serializer=neuroshard__pb2.DHTFindValueResponse.SerializeToString,
            ),
            'TensorExchange': grpc.unary_unary_rpc_method_handler(
                    servicer.TensorExchange,
                    request_deserializer=neuroshard__pb2.TensorExchangeRequest.FromString,
                    response_serializer=neuroshard__pb2.TensorExchangeResponse.SerializeToString,
            ),
            'SendPartialResult': grpc.unary_unary_rpc_method_handler(
                    servicer.SendPartialResult,
                    request_deserializer=neuroshard__pb2.PartialResultRequest.FromString,
                    response_serializer=neuroshard__pb2.PartialResultResponse.SerializeToString,
            ),
            'AnnounceTensorShard': grpc.unary_unary_rpc_method_handler(
                    servicer.AnnounceTensorShard,
                    request_deserializer=neuroshard__pb2.AnnounceShardRequest.FromString,
                    response_serializer=neuroshard__pb2.AnnounceShardResponse.SerializeToString,
            ),
            'FindTensorShardPeers': grpc.unary_unary_rpc_method_handler(
                    servicer.FindTensorShardPeers,
                    request_deserializer=neuroshard__pb2.FindShardPeersRequest.FromString,
                    response_serializer=neuroshard__pb2.FindShardPeersResponse.SerializeToString,
            ),
            'ListModels': grpc.unary_unary_rpc_method_handler(
                    servicer.ListModels,
                    request_deserializer=neuroshard__pb2.ListModelsRequest.FromString,
                    response_serializer=neuroshard__pb2.ListModelsResponse.SerializeToString,
            ),
            'GetModelStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetModelStatus,
                    request_deserializer=neuroshard__pb2.GetModelStatusRequest.FromString,
                    response_serializer=neuroshard__pb2.GetModelStatusResponse.SerializeToString,
            ),
            'SwarmForward': grpc.unary_unary_rpc_method_handler(
                    servicer.SwarmForward,
                    request_deserializer=neuroshard__pb2.SwarmForwardRequest.FromString,
                    response_serializer=neuroshard__pb2.SwarmForwardResponse.SerializeToString,
            ),
            'GetSwarmStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSwarmStatus,
                    request_deserializer=neuroshard__pb2.SwarmStatusRequest.FromString,
                    response_serializer=neuroshard__pb2.SwarmStatusResponse.SerializeToString,
            ),
            'UpdatePeerCapacity': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdatePeerCapacity,
                    request_deserializer=neuroshard__pb2.UpdatePeerCapacityRequest.FromString,
                    response_serializer=neuroshard__pb2.UpdatePeerCapacityResponse.SerializeToString,
            ),
            'AnnounceEpoch': grpc.unary_unary_rpc_method_handler(
                    servicer.AnnounceEpoch,
                    request_deserializer=neuroshard__pb2.EpochAnnouncementRequest.FromString,
                    response_serializer=neuroshard__pb2.EpochAnnouncementResponse.SerializeToString,
            ),
            'GetEpoch': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEpoch,
                    request_deserializer=neuroshard__pb2.EpochRequest.FromString,
                    response_serializer=neuroshard__pb2.EpochResponse.SerializeToString,
            ),
            'GetLatestEpoch': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLatestEpoch,
                    request_deserializer=neuroshard__pb2.Empty.FromString,
                    response_serializer=neuroshard__pb2.LatestEpochResponse.SerializeToString,
            ),
            'SpotCheckChallenge': grpc.unary_unary_rpc_method_handler(
                    servicer.SpotCheckChallenge,
                    request_deserializer=neuroshard__pb2.SpotCheckChallengeRequest.FromString,
                    response_serializer=neuroshard__pb2.SpotCheckChallengeResponse.SerializeToString,
            ),
            'ExpertForward': grpc.unary_unary_rpc_method_handler(
                    servicer.ExpertForward,
                    request_deserializer=neuroshard__pb2.ExpertForwardRequest.FromString,
                    response_serializer=neuroshard__pb2.ExpertForwardResponse.SerializeToString,
            ),
            'GetExpertStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetExpertStatus,
                    request_deserializer=neuroshard__pb2.ExpertStatusRequest.FromString,
                    response_serializer=neuroshard__pb2.ExpertStatusResponse.SerializeToString,
            ),
            'AnnounceExpert': grpc.unary_unary_rpc_method_handler(
                    servicer.AnnounceExpert,
                    request_deserializer=neuroshard__pb2.AnnounceExpertRequest.FromString,
                    response_serializer=neuroshard__pb2.AnnounceExpertResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'neuroshard.NeuroShardService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('neuroshard.NeuroShardService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class NeuroShardService(object):
    """Service definition for Node-to-Node and Client-to-Node communication
    """

    @staticmethod
    def StreamInference(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/neuroshard.NeuroShardService/StreamInference',
            neuroshard__pb2.InferenceRequest.SerializeToString,
            neuroshard__pb2.InferenceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UnaryInference(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/UnaryInference',
            neuroshard__pb2.InferenceRequest.SerializeToString,
            neuroshard__pb2.InferenceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetWeights(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetWeights',
            neuroshard__pb2.WeightRequest.SerializeToString,
            neuroshard__pb2.WeightResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GossipProof(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GossipProof',
            neuroshard__pb2.GossipProofRequest.SerializeToString,
            neuroshard__pb2.GossipProofResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GossipTransaction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GossipTransaction',
            neuroshard__pb2.GossipTransactionRequest.SerializeToString,
            neuroshard__pb2.GossipTransactionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GossipStake(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GossipStake',
            neuroshard__pb2.GossipStakeRequest.SerializeToString,
            neuroshard__pb2.GossipStakeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RequestProofValidation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/RequestProofValidation',
            neuroshard__pb2.ProofValidationRequest.SerializeToString,
            neuroshard__pb2.ProofValidationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GossipValidationVote(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GossipValidationVote',
            neuroshard__pb2.ValidationVoteRequest.SerializeToString,
            neuroshard__pb2.ValidationVoteResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GossipGradient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GossipGradient',
            neuroshard__pb2.GossipGradientRequest.SerializeToString,
            neuroshard__pb2.GossipGradientResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetCheckpoint(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetCheckpoint',
            neuroshard__pb2.GetCheckpointRequest.SerializeToString,
            neuroshard__pb2.GetCheckpointResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetCheckpointInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetCheckpointInfo',
            neuroshard__pb2.GetCheckpointInfoRequest.SerializeToString,
            neuroshard__pb2.GetCheckpointInfoResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PipelineForward(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/PipelineForward',
            neuroshard__pb2.PipelineForwardRequest.SerializeToString,
            neuroshard__pb2.PipelineForwardResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PipelineBackward(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/PipelineBackward',
            neuroshard__pb2.PipelineBackwardRequest.SerializeToString,
            neuroshard__pb2.PipelineBackwardResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetShardInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetShardInfo',
            neuroshard__pb2.GetShardInfoRequest.SerializeToString,
            neuroshard__pb2.GetShardInfoResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetShardChunk(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetShardChunk',
            neuroshard__pb2.GetShardChunkRequest.SerializeToString,
            neuroshard__pb2.GetShardChunkResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DHTPing(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/DHTPing',
            neuroshard__pb2.DHTPingRequest.SerializeToString,
            neuroshard__pb2.DHTPingResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DHTStore(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/DHTStore',
            neuroshard__pb2.DHTStoreRequest.SerializeToString,
            neuroshard__pb2.DHTStoreResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DHTFindNode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/DHTFindNode',
            neuroshard__pb2.DHTFindNodeRequest.SerializeToString,
            neuroshard__pb2.DHTFindNodeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DHTFindValue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/DHTFindValue',
            neuroshard__pb2.DHTFindValueRequest.SerializeToString,
            neuroshard__pb2.DHTFindValueResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TensorExchange(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/TensorExchange',
            neuroshard__pb2.TensorExchangeRequest.SerializeToString,
            neuroshard__pb2.TensorExchangeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendPartialResult(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/SendPartialResult',
            neuroshard__pb2.PartialResultRequest.SerializeToString,
            neuroshard__pb2.PartialResultResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AnnounceTensorShard(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/AnnounceTensorShard',
            neuroshard__pb2.AnnounceShardRequest.SerializeToString,
            neuroshard__pb2.AnnounceShardResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def FindTensorShardPeers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/FindTensorShardPeers',
            neuroshard__pb2.FindShardPeersRequest.SerializeToString,
            neuroshard__pb2.FindShardPeersResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListModels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/ListModels',
            neuroshard__pb2.ListModelsRequest.SerializeToString,
            neuroshard__pb2.ListModelsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetModelStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetModelStatus',
            neuroshard__pb2.GetModelStatusRequest.SerializeToString,
            neuroshard__pb2.GetModelStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SwarmForward(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/SwarmForward',
            neuroshard__pb2.SwarmForwardRequest.SerializeToString,
            neuroshard__pb2.SwarmForwardResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetSwarmStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetSwarmStatus',
            neuroshard__pb2.SwarmStatusRequest.SerializeToString,
            neuroshard__pb2.SwarmStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpdatePeerCapacity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/UpdatePeerCapacity',
            neuroshard__pb2.UpdatePeerCapacityRequest.SerializeToString,
            neuroshard__pb2.UpdatePeerCapacityResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AnnounceEpoch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/AnnounceEpoch',
            neuroshard__pb2.EpochAnnouncementRequest.SerializeToString,
            neuroshard__pb2.EpochAnnouncementResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetEpoch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetEpoch',
            neuroshard__pb2.EpochRequest.SerializeToString,
            neuroshard__pb2.EpochResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetLatestEpoch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetLatestEpoch',
            neuroshard__pb2.Empty.SerializeToString,
            neuroshard__pb2.LatestEpochResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SpotCheckChallenge(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/SpotCheckChallenge',
            neuroshard__pb2.SpotCheckChallengeRequest.SerializeToString,
            neuroshard__pb2.SpotCheckChallengeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ExpertForward(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/ExpertForward',
            neuroshard__pb2.ExpertForwardRequest.SerializeToString,
            neuroshard__pb2.ExpertForwardResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetExpertStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/GetExpertStatus',
            neuroshard__pb2.ExpertStatusRequest.SerializeToString,
            neuroshard__pb2.ExpertStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AnnounceExpert(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/neuroshard.NeuroShardService/AnnounceExpert',
            neuroshard__pb2.AnnounceExpertRequest.SerializeToString,
            neuroshard__pb2.AnnounceExpertResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""
P2P DataSwarm — Decentralized Shard Distribution for NeuroShard

Implements BitTorrent-style P2P data transfer so nodes can fetch training
data shards from peers instead of always hitting the CDN. This removes
the CDN as a single point of failure and distributes bandwidth costs.

Architecture:
1. Nodes announce which shards they hold to the DHT (every 60s)
2. When a node needs a shard, it queries the DHT for providers
3. If P2P providers exist, download chunks in parallel from peers
4. If no peers or P2P fails, fall back to CDN (CloudFront)
5. After downloading, announce the new shard to DHT

Chunk Protocol:
- Shards are split into 1MB chunks
- Each chunk is requested via gRPC DataChunk RPC
- Chunks are reassembled and verified before loading
"""

import os
import hashlib
import logging
import threading
import time
import requests
import math
from typing import List, Dict, Optional, Set
from concurrent.futures import ThreadPoolExecutor, as_completed

logger = logging.getLogger(__name__)

# Avoid circular imports — P2PManager type is used for type hints only
try:
    from neuroshard.core.network.p2p import P2PManager, GRPC_PORT_OFFSET
except ImportError:
    P2PManager = None
    GRPC_PORT_OFFSET = 1000


class DataSwarm:
    """
    Implements BitTorrent-like P2P data transfer for NeuroShard.
    
    - Splits large shards into 1MB chunks.
    - Finds peers holding specific shards via DHT.
    - Downloads chunks in parallel from multiple peers.
    - Falls back to CDN when no peers are available.
    - Verifies data integrity after download.
    """
    
    CHUNK_SIZE = 1024 * 1024  # 1MB chunks
    ANNOUNCE_INTERVAL = 60   # Announce shards to DHT every 60s
    
    def __init__(self, p2p_manager, cache_dir: str = "data_cache"):
        self.p2p = p2p_manager
        self.cache_dir = cache_dir
        self.active_downloads: Dict[int, str] = {}  # shard_id -> status
        self.local_shards: Set[int] = set()  # IDs of shards we have locally
        
        # Thread pool for parallel chunk downloads
        self.executor = ThreadPoolExecutor(max_workers=8, thread_name_prefix="swarm-dl")
        
        os.makedirs(cache_dir, exist_ok=True)
        self._scan_local_cache()
        
        # Start announcer thread (tells DHT which shards we have)
        self._announce_thread = threading.Thread(
            target=self._announce_loop, daemon=True, name="shard-announce"
        )
        self._announce_thread.start()
        
    def _scan_local_cache(self):
        """Scan cache directory for existing complete shards."""
        try:
            for filename in os.listdir(self.cache_dir):
                if filename.startswith("genesis_shard_") and filename.endswith(".pt"):
                    try:
                        idx = int(filename.split("_")[2].split(".")[0])
                        self.local_shards.add(idx)
                    except (ValueError, IndexError):
                        pass
            if self.local_shards:
                logger.info(f"[SWARM] Found {len(self.local_shards)} cached shards locally")
        except OSError as e:
            logger.debug(f"[SWARM] Cache scan error: {e}")

    def _announce_loop(self):
        """Periodically announce our shards to the DHT so peers can find us."""
        while True:
            try:
                if self.p2p and hasattr(self.p2p, 'dht') and self.p2p.dht:
                    my_endpoint = getattr(self.p2p, 'my_url', None) or ''
                    if my_endpoint:
                        for shard_id in list(self.local_shards):
                            try:
                                dht_key = f"shard_provider_{shard_id}"
                                # Store our endpoint as the provider for this shard
                                self.p2p.dht.store(dht_key, my_endpoint)
                            except Exception as e:
                                logger.debug(f"[SWARM] DHT announce failed for shard {shard_id}: {e}")
            except Exception as e:
                logger.debug(f"[SWARM] Announce loop error: {e}")
            time.sleep(self.ANNOUNCE_INTERVAL)

    def get_shard_path(self, shard_id: int) -> str:
        return os.path.join(self.cache_dir, f"genesis_shard_{shard_id}.pt")

    def download_shard(self, shard_id: int, manifest_url: str = None) -> str:
        """
        Download a shard using P2P swarm, falling back to CDN.
        Returns path to downloaded file.
        """
        target_path = self.get_shard_path(shard_id)
        
        if shard_id in self.local_shards and os.path.exists(target_path):
            return target_path
            
        logger.info(f"[SWARM] Starting download for shard {shard_id}...")
        
        # 1. Try P2P: Find peers who have this shard via DHT
        peers = self._find_providers(shard_id)
        
        if peers:
            logger.info(f"[SWARM] Found {len(peers)} P2P providers for shard {shard_id}")
            success = self._swarm_download(shard_id, peers, target_path)
            if success:
                self.local_shards.add(shard_id)
                return target_path
            logger.warning(f"[SWARM] P2P download failed for shard {shard_id}, falling back to CDN")
        
        # 2. Fallback to CDN
        return self._download_from_cdn(shard_id, target_path, manifest_url)

    def _find_providers(self, shard_id: int) -> List[str]:
        """Find peers that have this shard by querying the DHT.
        
        Returns list of peer URLs (e.g., ['http://1.2.3.4:8000', ...]).
        """
        providers = []
        
        if not self.p2p or not hasattr(self.p2p, 'dht') or not self.p2p.dht:
            return providers
        
        try:
            dht_key = f"shard_provider_{shard_id}"
            
            # Query DHT for providers
            result = self.p2p.dht.lookup_value(dht_key)
            if result:
                if isinstance(result, str):
                    # Single provider
                    providers.append(result)
                elif isinstance(result, list):
                    providers.extend(result)
            
            # Also check known peers — ask them directly if they have it
            for peer_url in list(self.p2p.known_peers.keys())[:10]:
                if peer_url not in providers:
                    if self._peer_has_shard(peer_url, shard_id):
                        providers.append(peer_url)
                        
        except Exception as e:
            logger.debug(f"[SWARM] Provider discovery error: {e}")
        
        # Filter out self
        my_url = getattr(self.p2p, 'my_url', None)
        return [p for p in providers if p != my_url]
    
    def _peer_has_shard(self, peer_url: str, shard_id: int) -> bool:
        """Check if a peer has a specific shard via gRPC."""
        try:
            from neuroshard.core.network.connection_pool import get_channel
            from neuroshard.protos import neuroshard_pb2 as pb2
            from neuroshard.protos import neuroshard_pb2_grpc as pb2_grpc
            from urllib.parse import urlparse
            
            parsed = urlparse(peer_url)
            grpc_addr = f"{parsed.hostname}:{(parsed.port or 80) + GRPC_PORT_OFFSET}"
            channel = get_channel(grpc_addr)
            stub = pb2_grpc.NeuroShardStub(channel)
            
            # Use GetStatus to check if peer has this shard
            resp = stub.GetStatus(pb2.Empty(), timeout=3.0)
            # Check if shard is in peer's data list
            if hasattr(resp, 'available_shards'):
                return shard_id in resp.available_shards
        except Exception:
            pass
        return False

    def _download_from_cdn(self, shard_id: int, target_path: str, manifest_url: str) -> str:
        """Download from CloudFront CDN (single attempt — caller handles retry)."""
        url = manifest_url or f"https://dwquwt9gkkeil.cloudfront.net/shard_{shard_id}.pt"
        
        try:
            with requests.get(url, stream=True, timeout=60) as r:
                r.raise_for_status()
                tmp_path = target_path + f".tmp.{os.getpid()}"
                with open(tmp_path, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=8192): 
                        f.write(chunk)
                os.replace(tmp_path, target_path)
            self.local_shards.add(shard_id)
            logger.info(f"[SWARM] Downloaded shard {shard_id} from CDN")
            return target_path
        except Exception as e:
            raise RuntimeError(f"Failed to download shard {shard_id} from {url}: {e}")

    def _swarm_download(self, shard_id: int, peers: List[str], target_path: str) -> bool:
        """Download shard chunks in parallel from multiple peers.
        
        1. Get file size from first responding peer
        2. Split into chunks, assign to peers round-robin
        3. Download all chunks in parallel
        4. Reassemble and verify
        """
        try:
            # Get shard metadata from first available peer
            file_size = self._get_shard_size_from_peer(peers[0], shard_id)
            if file_size is None or file_size == 0:
                return False
            
            num_chunks = math.ceil(file_size / self.CHUNK_SIZE)
            logger.info(f"[SWARM] Shard {shard_id}: {file_size/1e6:.1f}MB, {num_chunks} chunks, "
                       f"{len(peers)} peers")
            
            # Prepare chunk assignments (round-robin across peers)
            chunk_data = [None] * num_chunks
            futures = {}
            
            for chunk_idx in range(num_chunks):
                peer = peers[chunk_idx % len(peers)]
                future = self.executor.submit(
                    self._download_chunk_from_peer, peer, shard_id, chunk_idx
                )
                futures[future] = chunk_idx
            
            # Collect results
            failed_chunks = []
            for future in as_completed(futures, timeout=120):
                chunk_idx = futures[future]
                try:
                    data = future.result()
                    if data is not None:
                        chunk_data[chunk_idx] = data
                    else:
                        failed_chunks.append(chunk_idx)
                except Exception as e:
                    logger.debug(f"[SWARM] Chunk {chunk_idx} failed: {e}")
                    failed_chunks.append(chunk_idx)
            
            # Retry failed chunks from CDN
            if failed_chunks:
                logger.warning(f"[SWARM] {len(failed_chunks)}/{num_chunks} chunks failed from peers, "
                              f"retrying from CDN")
                for chunk_idx in failed_chunks:
                    data = self._download_chunk_from_cdn(shard_id, chunk_idx)
                    if data is not None:
                        chunk_data[chunk_idx] = data
                    else:
                        logger.error(f"[SWARM] Chunk {chunk_idx} unrecoverable")
                        return False
            
            # Reassemble
            if any(c is None for c in chunk_data):
                return False
            
            tmp_path = target_path + f".tmp.{os.getpid()}"
            with open(tmp_path, 'wb') as f:
                for chunk in chunk_data:
                    f.write(chunk)
            os.replace(tmp_path, target_path)
            
            logger.info(f"[SWARM] Shard {shard_id} assembled from {len(peers)} peers "
                       f"({num_chunks} chunks)")
            return True
            
        except Exception as e:
            logger.warning(f"[SWARM] Swarm download failed for shard {shard_id}: {e}")
            return False
    
    def _get_shard_size_from_peer(self, peer_url: str, shard_id: int) -> Optional[int]:
        """Get shard file size from a peer."""
        try:
            from neuroshard.core.network.connection_pool import get_channel
            from neuroshard.protos import neuroshard_pb2 as pb2
            from neuroshard.protos import neuroshard_pb2_grpc as pb2_grpc
            from urllib.parse import urlparse
            
            parsed = urlparse(peer_url)
            grpc_addr = f"{parsed.hostname}:{(parsed.port or 80) + GRPC_PORT_OFFSET}"
            channel = get_channel(grpc_addr)
            stub = pb2_grpc.NeuroShardStub(channel)
            
            request = pb2.DataChunkRequest(
                shard_id=shard_id,
                chunk_index=-1,  # -1 = metadata request
            )
            response = stub.DataChunk(request, timeout=10.0)
            return response.total_size if hasattr(response, 'total_size') else None
        except Exception as e:
            logger.debug(f"[SWARM] Metadata request to {peer_url} failed: {e}")
            return None
    
    def _download_chunk_from_peer(self, peer_url: str, shard_id: int, chunk_index: int) -> Optional[bytes]:
        """Download a single chunk from a peer via gRPC."""
        try:
            from neuroshard.core.network.connection_pool import get_channel
            from neuroshard.protos import neuroshard_pb2 as pb2
            from neuroshard.protos import neuroshard_pb2_grpc as pb2_grpc
            from urllib.parse import urlparse
            
            parsed = urlparse(peer_url)
            grpc_addr = f"{parsed.hostname}:{(parsed.port or 80) + GRPC_PORT_OFFSET}"
            channel = get_channel(grpc_addr)
            stub = pb2_grpc.NeuroShardStub(channel)
            
            request = pb2.DataChunkRequest(
                shard_id=shard_id,
                chunk_index=chunk_index,
            )
            response = stub.DataChunk(request, timeout=30.0)
            
            if response.data and len(response.data) > 0:
                return response.data
            return None
        except Exception as e:
            logger.debug(f"[SWARM] Chunk {chunk_index} from {peer_url} failed: {e}")
            return None
    
    def _download_chunk_from_cdn(self, shard_id: int, chunk_index: int) -> Optional[bytes]:
        """Download a single chunk range from CDN (byte-range request)."""
        try:
            url = f"https://dwquwt9gkkeil.cloudfront.net/shard_{shard_id}.pt"
            start = chunk_index * self.CHUNK_SIZE
            end = start + self.CHUNK_SIZE - 1
            headers = {'Range': f'bytes={start}-{end}'}
            
            resp = requests.get(url, headers=headers, timeout=30)
            if resp.status_code in (200, 206):
                return resp.content
            return None
        except Exception:
            return None

    def serve_chunk(self, shard_id: int, chunk_index: int) -> Optional[bytes]:
        """Read a chunk from disk to serve to a peer.
        
        Called by the gRPC DataChunk handler when another node requests data.
        """
        if shard_id not in self.local_shards:
            return None
            
        path = self.get_shard_path(shard_id)
        if not os.path.exists(path):
            return None
        
        offset = chunk_index * self.CHUNK_SIZE
        
        try:
            with open(path, "rb") as f:
                f.seek(offset)
                data = f.read(self.CHUNK_SIZE)
                return data if data else None
        except (OSError, IOError) as e:
            logger.debug(f"[SWARM] Failed to read chunk {chunk_index} of shard {shard_id}: {e}")
            return None
    
    def get_shard_file_size(self, shard_id: int) -> int:
        """Get the file size of a local shard (for metadata responses)."""
        path = self.get_shard_path(shard_id)
        try:
            return os.path.getsize(path) if os.path.exists(path) else 0
        except OSError:
            return 0



"""
NEP Voting - Stake-Weighted Governance

Voting power is proportional to staked NEURO. This ensures that
participants with skin in the game make decisions about protocol changes.

Voting Rules:
- 1 NEURO staked = 1 vote
- Voting period: 7 days (configurable per NEP)
- Approval threshold: 66% of votes (stake-weighted)
- Quorum: 20% of total staked NEURO must vote
"""

import sqlite3
import time
import hashlib
import json
import logging
import threading
from dataclasses import dataclass, asdict
from enum import Enum
from typing import Dict, List, Optional, Tuple

logger = logging.getLogger(__name__)


class VoteChoice(Enum):
    """Voting options."""
    YES = "yes"           # Approve the proposal
    NO = "no"             # Reject the proposal
    ABSTAIN = "abstain"   # Participate without opinion (counts for quorum)


@dataclass
class Vote:
    """A single vote on a NEP."""
    vote_id: str
    nep_id: str
    voter_node_id: str
    choice: VoteChoice
    stake_at_vote: float      # Stake when vote was cast (locked)
    timestamp: float
    signature: str            # ECDSA signature
    reason: str = ""          # Optional reason for vote
    
    def to_dict(self) -> Dict:
        return {
            "vote_id": self.vote_id,
            "nep_id": self.nep_id,
            "voter_node_id": self.voter_node_id,
            "choice": self.choice.value,
            "stake_at_vote": self.stake_at_vote,
            "timestamp": self.timestamp,
            "signature": self.signature,
            "reason": self.reason,
        }
    
    @classmethod
    def from_dict(cls, data: Dict) -> 'Vote':
        return cls(
            vote_id=data["vote_id"],
            nep_id=data["nep_id"],
            voter_node_id=data["voter_node_id"],
            choice=VoteChoice(data["choice"]),
            stake_at_vote=data["stake_at_vote"],
            timestamp=data["timestamp"],
            signature=data["signature"],
            reason=data.get("reason", ""),
        )


@dataclass
class VoteResult:
    """Aggregated voting results for a NEP."""
    nep_id: str
    
    # Vote counts (number of voters)
    yes_count: int = 0
    no_count: int = 0
    abstain_count: int = 0
    
    # Stake-weighted votes
    yes_stake: float = 0.0
    no_stake: float = 0.0
    abstain_stake: float = 0.0
    
    # Thresholds
    total_stake_voted: float = 0.0
    total_network_stake: float = 0.0
    approval_threshold: float = 0.66
    quorum_threshold: float = 0.20
    
    # Status
    quorum_reached: bool = False
    approved: bool = False
    
    @property
    def participation_rate(self) -> float:
        """Percentage of network stake that voted."""
        if self.total_network_stake == 0:
            return 0.0
        return self.total_stake_voted / self.total_network_stake
    
    @property
    def approval_rate(self) -> float:
        """Percentage of voting stake that approved."""
        yes_no_total = self.yes_stake + self.no_stake
        if yes_no_total == 0:
            return 0.0
        return self.yes_stake / yes_no_total
    
    def to_dict(self) -> Dict:
        return {
            "nep_id": self.nep_id,
            "yes_count": self.yes_count,
            "no_count": self.no_count,
            "abstain_count": self.abstain_count,
            "yes_stake": self.yes_stake,
            "no_stake": self.no_stake,
            "abstain_stake": self.abstain_stake,
            "total_stake_voted": self.total_stake_voted,
            "total_network_stake": self.total_network_stake,
            "participation_rate": self.participation_rate,
            "approval_rate": self.approval_rate,
            "quorum_reached": self.quorum_reached,
            "approved": self.approved,
        }


class VotingSystem:
    """
    Manages NEP voting with stake-weighted governance.
    """
    
    def __init__(
        self,
        db_path: str = "nep_votes.db",
        ledger=None,  # NEUROLedger for stake lookups
    ):
        self.db_path = db_path
        self.ledger = ledger
        self.lock = threading.Lock()
        
        self._init_db()
    
    def _init_db(self):
        """Initialize voting database."""
        with self.lock:
            with sqlite3.connect(self.db_path, timeout=60.0) as conn:
                conn.execute("""
                    CREATE TABLE IF NOT EXISTS votes (
                        vote_id TEXT PRIMARY KEY,
                        nep_id TEXT NOT NULL,
                        voter_node_id TEXT NOT NULL,
                        choice TEXT NOT NULL,
                        stake_at_vote REAL NOT NULL,
                        timestamp REAL NOT NULL,
                        signature TEXT NOT NULL,
                        reason TEXT DEFAULT '',
                        UNIQUE(nep_id, voter_node_id)
                    )
                """)
                
                conn.execute("""
                    CREATE INDEX IF NOT EXISTS idx_votes_nep 
                    ON votes(nep_id)
                """)
    
    def cast_vote(
        self,
        nep_id: str,
        voter_node_id: str,
        choice: VoteChoice,
        signature: str,
        reason: str = "",
        nep_registry=None,
    ) -> Tuple[bool, str]:
        """
        Cast a vote on a NEP.
        
        Requirements:
        1. NEP must be in VOTING status
        2. Voting period must be active
        3. Voter must have stake > 0
        4. Vote must be signed
        
        Returns: (success, message)
        """
        # Check NEP status
        if nep_registry:
            nep = nep_registry.get_proposal(nep_id)
            if not nep:
                return False, "NEP not found"
            if not nep.is_voting_active():
                return False, f"Voting not active for {nep_id}"
        
        # Get voter's stake
        stake = 0.0
        if self.ledger:
            stake = self.ledger.get_local_stake(voter_node_id)
            if stake <= 0:
                return False, "Must have staked NEURO to vote"
        else:
            # Test mode - assume 100 stake
            stake = 100.0
        
        # Verify ECDSA signature before accepting the vote
        try:
            from neuroshard.core.crypto.ecdsa import verify_signature
            vote_payload = f"{nep_id}:{voter_node_id}:{choice.value}:{reason}"
            if not verify_signature(voter_node_id, vote_payload, signature):
                return False, "Invalid vote signature — ECDSA verification failed"
        except ImportError:
            logger.warning("ECDSA module unavailable — skipping vote signature verification")
        except Exception as e:
            logger.warning(f"Vote signature verification error: {e}")
            return False, f"Signature verification error: {e}"
        
        # Check if already voted
        with sqlite3.connect(self.db_path, timeout=60.0) as conn:
            existing = conn.execute(
                "SELECT vote_id FROM votes WHERE nep_id = ? AND voter_node_id = ?",
                (nep_id, voter_node_id)
            ).fetchone()
            
            if existing:
                return False, "Already voted on this NEP"
        
        # Create vote
        vote_id = hashlib.sha256(
            f"{nep_id}:{voter_node_id}:{time.time()}".encode()
        ).hexdigest()[:32]
        
        vote = Vote(
            vote_id=vote_id,
            nep_id=nep_id,
            voter_node_id=voter_node_id,
            choice=choice,
            stake_at_vote=stake,
            timestamp=time.time(),
            signature=signature,
            reason=reason,
        )
        
        # Store vote
        with self.lock:
            with sqlite3.connect(self.db_path, timeout=60.0) as conn:
                conn.execute("""
                    INSERT INTO votes 
                    (vote_id, nep_id, voter_node_id, choice, stake_at_vote, 
                     timestamp, signature, reason)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                """, (
                    vote.vote_id,
                    vote.nep_id,
                    vote.voter_node_id,
                    vote.choice.value,
                    vote.stake_at_vote,
                    vote.timestamp,
                    vote.signature,
                    vote.reason,
                ))
        
        logger.info(
            f"Vote cast: {voter_node_id[:16]}... voted {choice.value} "
            f"on {nep_id} with {stake:.2f} stake"
        )
        
        return True, f"Vote recorded: {choice.value} with {stake:.2f} stake"
    
    def get_vote_tally(
        self,
        nep_id: str,
        approval_threshold: float = 0.66,
        quorum_threshold: float = 0.20,
    ) -> VoteResult:
        """
        Get the current vote tally for a NEP.
        """
        result = VoteResult(
            nep_id=nep_id,
            approval_threshold=approval_threshold,
            quorum_threshold=quorum_threshold,
        )
        
        with sqlite3.connect(self.db_path, timeout=60.0) as conn:
            votes = conn.execute(
                "SELECT choice, stake_at_vote FROM votes WHERE nep_id = ?",
                (nep_id,)
            ).fetchall()
            
            for choice_str, stake in votes:
                choice = VoteChoice(choice_str)
                
                if choice == VoteChoice.YES:
                    result.yes_count += 1
                    result.yes_stake += stake
                elif choice == VoteChoice.NO:
                    result.no_count += 1
                    result.no_stake += stake
                else:
                    result.abstain_count += 1
                    result.abstain_stake += stake
        
        result.total_stake_voted = (
            result.yes_stake + result.no_stake + result.abstain_stake
        )
        
        # Get total network stake
        if self.ledger:
            # Query total staked from ledger
            result.total_network_stake = self._get_total_network_stake()
        else:
            # Test mode - assume voted stake is 50% of network
            result.total_network_stake = result.total_stake_voted * 2
        
        # Check quorum
        if result.total_network_stake > 0:
            result.quorum_reached = (
                result.participation_rate >= quorum_threshold
            )
        
        # Check approval (only if quorum reached)
        if result.quorum_reached:
            result.approved = result.approval_rate >= approval_threshold
        
        return result
    
    def _get_total_network_stake(self) -> float:
        """Get total staked NEURO in the network."""
        if not self.ledger:
            return 0.0
        
        try:
            with sqlite3.connect(self.ledger.db_path, timeout=60.0) as conn:
                row = conn.execute(
                    "SELECT SUM(amount) FROM stakes WHERE locked_until > ?",
                    (time.time(),)
                ).fetchone()
                return row[0] if row and row[0] else 0.0
        except Exception as e:
            logger.error(f"Failed to get network stake: {e}")
            return 0.0
    
    def get_voter_history(self, voter_node_id: str) -> List[Vote]:
        """Get all votes by a node."""
        with sqlite3.connect(self.db_path, timeout=60.0) as conn:
            rows = conn.execute(
                "SELECT vote_id, nep_id, voter_node_id, choice, "
                "stake_at_vote, timestamp, signature, reason "
                "FROM votes WHERE voter_node_id = ? ORDER BY timestamp DESC",
                (voter_node_id,)
            ).fetchall()
            
            return [
                Vote(
                    vote_id=row[0],
                    nep_id=row[1],
                    voter_node_id=row[2],
                    choice=VoteChoice(row[3]),
                    stake_at_vote=row[4],
                    timestamp=row[5],
                    signature=row[6],
                    reason=row[7],
                )
                for row in rows
            ]
    
    def finalize_voting(
        self,
        nep_id: str,
        nep_registry,
    ) -> Tuple[bool, str, VoteResult]:
        """
        Finalize voting and update NEP status.
        
        Called when voting period ends.
        Returns: (success, message, final_result)
        """
        nep = nep_registry.get_proposal(nep_id)
        if not nep:
            return False, "NEP not found", None
        
        # Get final tally
        result = self.get_vote_tally(
            nep_id,
            approval_threshold=nep.approval_threshold,
            quorum_threshold=nep.quorum_threshold,
        )
        
        if result.approved:
            # Calculate activation block (e.g., 10000 blocks from now)
            activation_block = self._get_current_block() + nep.upgrade_path.activation_delay_blocks
            
            nep_registry.update_status(
                nep_id,
                NEPStatus.SCHEDULED,
                activation_block=activation_block,
            )
            
            message = (
                f"Approved! Activation scheduled at block {activation_block}. "
                f"Approval: {result.approval_rate:.1%}, "
                f"Participation: {result.participation_rate:.1%}"
            )
            
        elif result.quorum_reached:
            nep_registry.update_status(nep_id, NEPStatus.REJECTED)
            message = (
                f"Rejected. Approval: {result.approval_rate:.1%} "
                f"(needed {nep.approval_threshold:.1%})"
            )
            
        else:
            nep_registry.update_status(nep_id, NEPStatus.REJECTED)
            message = (
                f"Failed to reach quorum. Participation: {result.participation_rate:.1%} "
                f"(needed {nep.quorum_threshold:.1%})"
            )
        
        logger.info(f"Voting finalized for {nep_id}: {message}")
        
        return True, message, result
    
    def _get_current_block(self) -> int:
        """
        Get current block number.
        
        In a true blockchain, this would query the chain.
        For NeuroShard, we use a logical block based on proof count.
        """
        if self.ledger:
            try:
                stats = self.ledger.get_global_stats()
                # Use proof count as "block height"
                return stats.total_proofs_processed
            except:
                pass
        return int(time.time() / 60)  # Fallback: 1 "block" per minute


# Module-level convenience functions
from .proposal import NEPStatus

def cast_vote(
    voting_system: VotingSystem,
    nep_id: str,
    voter_node_id: str,
    choice: VoteChoice,
    signature: str,
    reason: str = "",
    nep_registry=None,
) -> Tuple[bool, str]:
    """Cast a vote on a NEP."""
    return voting_system.cast_vote(
        nep_id, voter_node_id, choice, signature, reason, nep_registry
    )


def get_vote_tally(
    voting_system: VotingSystem,
    nep_id: str,
) -> VoteResult:
    """Get current vote tally for a NEP."""
    return voting_system.get_vote_tally(nep_id)
